import os
import struct
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

class MessageHeader:
    """
    Represents the header of an encrypted message.
      - flag (16 bits)
      - data_len (16 bits): length in bytes of the encrypted payload (excluding tag)
      - Associated Data (AD):
          * retry (8 bits)
          * connexion_status (4 bits) + 4 bits padding (packed in one byte)
          * iv/messageID (96 bits / 12 bytes)
    Total header size: 2 + 2 + 1 + 1 + 12 = 18 bytes.
    """
    def __init__(self, flag: int, data_len: int, retry: int, connexion_status: int, iv: bytes):
        self.flag = flag              # 16 bits
        self.data_len = data_len      # 16 bits
        self.retry = retry            # 8 bits
        self.connexion_status = connexion_status  # 4 bits
        self.iv = iv                  # 96 bits (12 bytes)

    def pack(self) -> bytes:
        # Pack flag and data_len as unsigned shorts (2 bytes each)
        header = struct.pack('>H H', self.flag, self.data_len)
        # Pack retry (1 byte) and connexion_status (4 bits in high nibble, 4 bits padding as zero)
        ad_byte = (self.connexion_status & 0x0F) << 4
        ad_packed = struct.pack('>B B', self.retry, ad_byte)
        # Append IV (12 bytes)
        return header + ad_packed + self.iv

    @classmethod
    def unpack(cls, data: bytes) -> 'MessageHeader':
        # Expect exactly 18 bytes
        flag, data_len = struct.unpack('>H H', data[:4])
        retry, ad_byte = struct.unpack('>B B', data[4:6])
        connexion_status = (ad_byte >> 4) & 0x0F
        iv = data[6:18]
        return cls(flag, data_len, retry, connexion_status, iv)

def generate_iv(initial: bool, previous_iv: bytes = None) -> bytes:
    """
    Generate a 96-bit IV (12 bytes).
      - If 'initial' is True, return a random IV.
      - Otherwise, increment the previous IV by 1 modulo 2^96.
    """
    if initial or previous_iv is None:
        return os.urandom(12)
    else:
        iv_int = int.from_bytes(previous_iv, 'big')
        iv_int = (iv_int + 1) % (1 << 96)
        return iv_int.to_bytes(12, 'big')

def encrypt_message(plaintext: bytes, key: bytes, flag: int = 0xBEEF, retry: int = 0, connexion_status: int = 0) -> bytes:
    """
    Encrypts a plaintext using AES-256-GCM.
      - Generates a random 96-bit IV.
      - Encrypts the plaintext with AESGCM.
      - Builds a MessageHeader with the provided flag, the data_len (length of ciphertext excluding tag),
        retry, connexion_status, and the IV.
      - Returns the full encrypted message: header (18 bytes) || ciphertext || tag (16 bytes).
    """
    aesgcm = AESGCM(key)
    iv = generate_iv(initial=True)
    # Encrypt with no associated data (you may later use the header as AD if needed)
    ciphertext_with_tag = aesgcm.encrypt(iv, plaintext, None)
    tag_length = 16  # default tag size
    ciphertext = ciphertext_with_tag[:-tag_length]
    tag = ciphertext_with_tag[-tag_length:]
    data_len = len(ciphertext)
    header = MessageHeader(flag=flag, data_len=data_len, retry=retry, connexion_status=connexion_status, iv=iv)
    packed_header = header.pack()
    return packed_header + ciphertext + tag

def decrypt_message(message: bytes, key: bytes) -> bytes:
    """
    Decrypts a message that was encrypted with encrypt_message.
    Expects message format: header (18 bytes) || ciphertext || tag (16 bytes).
    Returns the decrypted plaintext.
    """
    if len(message) < 18 + 16:
        raise ValueError("Message too short.")
    header_bytes = message[:18]
    header = MessageHeader.unpack(header_bytes)
    data_len = header.data_len
    expected_len = 18 + data_len + 16
    if len(message) != expected_len:
        raise ValueError("Message length does not match header's data_len.")
    ciphertext = message[18:18+data_len]
    tag = message[18+data_len:]
    ciphertext_with_tag = ciphertext + tag
    aesgcm = AESGCM(key)
    plaintext = aesgcm.decrypt(header.iv, ciphertext_with_tag, None)
    return plaintext
