import random
import os
import time
import threading
from typing import List, Dict, Any

from crypto_utils import (
    generate_identity_keys,
    load_peer_identity_key,
    sign_data,
    verify_signature,
    get_ephemeral_keypair,
    compute_ecdh_shared_key
)
from messages import (
    build_ping_request, parse_ping_request,
    build_ping_response, parse_ping_response,
    build_handshake_message, parse_handshake_message,
    compute_pfs_hash
)
import transmission

from encryption import encrypt_message, decrypt_message, MessageHeader, generate_iv

# ANSI colors
RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"


class IcingProtocol:
    def __init__(self):
        # Identity keys (each 512 bits when printed as hex of 64 bytes)
        self.identity_privkey, self.identity_pubkey = generate_identity_keys()

        # Peer identity for verifying ephemeral signatures
        self.peer_identity_pubkey_obj = None
        self.peer_identity_pubkey_bytes = None

        # Ephemeral keys (our side)
        self.ephemeral_privkey = None
        self.ephemeral_pubkey = None

        # Last computed shared secret (hex string)
        self.shared_secret = None

        # Derived HKDF key (hex string, 256 bits)
        self.hkdf_key = None

        # For PFS: track per-peer session info (session number and last shared secret)
        self.pfs_history: Dict[bytes, (int, str)] = {}

        # Protocol flags
        self.state = {
            "ping_sent": False,
            "ping_received": False,
            "handshake_sent": False,
            "handshake_received": False,
        }

        # Auto-responder toggle
        self.auto_responder = False

        # Active connections list
        self.connections = []

        # Inbound messages (each message is a dict with keys: type, raw, parsed, connection)
        self.inbound_messages: List[Dict[str, Any]] = []

        # Store the session nonce (32 bytes) from our first sent or received PING
        self.session_nonce: bytes = None

        self.local_port = random.randint(30000, 40000)
        self.server_listener = transmission.ServerListener(
            host="127.0.0.1",
            port=self.local_port,
            on_new_connection=self.on_new_connection,
            on_data_received=self.on_data_received
        )
        self.server_listener.start()

    # -------------------------------------------------------------------------
    # Transport callbacks
    # -------------------------------------------------------------------------

    def on_new_connection(self, conn: transmission.PeerConnection):
        print(f"{GREEN}[IcingProtocol]{RESET} New incoming connection.")
        self.connections.append(conn)

    def on_data_received(self, conn: transmission.PeerConnection, data: bytes):
        bits_count = len(data) * 8
        print(
            f"{GREEN}[RECV]{RESET} {bits_count} bits from peer: {data.hex()[:60]}{'...' if len(data.hex()) > 60 else ''}")

        # New-format PING REQUEST (22 bytes)
        if len(data) == 22:
            parsed = parse_ping_request(data)
            if parsed:
                nonce_full, version, cipher = parsed
                self.state["ping_received"] = True
                # If the received cipher field is not 0, we force it to 0 in our response.
                if cipher != 0:
                    print(
                        f"{YELLOW}[NOTICE]{RESET} Received PING with unsupported cipher ({cipher}); forcing cipher to 0 in response.")
                    cipher = 0
                # Store session nonce if not already set:
                if self.session_nonce is None:
                    # Here, we already generated 17 bytes (136 bits) and only the top 129 bits are valid.
                    nonce_int = int.from_bytes(nonce_full, 'big') >> 7
                    self.session_nonce = nonce_int.to_bytes(17, 'big')
                    print(f"{YELLOW}[NOTICE]{RESET} Stored session nonce from received PING.")
                index = len(self.inbound_messages)
                msg = {
                    "type": "PING_REQUEST",
                    "raw": data,
                    "parsed": {"nonce_full": nonce_full, "version": version, "cipher": cipher},
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                # (Optional auto-responder code could go here)
                return

        # New-format PING RESPONSE (10 bytes)
        elif len(data) == 10:
            parsed = parse_ping_response(data)
            if parsed:
                timestamp, version, cipher, answer = parsed
                # If cipher is not 0 (AES-256-GCM), override it.
                if cipher != 0:
                    print(
                        f"{YELLOW}[NOTICE]{RESET} Received PING RESPONSE with unsupported cipher; treating as AES (cipher=0).")
                    cipher = 0
                index = len(self.inbound_messages)
                msg = {
                    "type": "PING_RESPONSE",
                    "raw": data,
                    "parsed": {"timestamp": timestamp, "version": version, "cipher": cipher, "answer": answer},
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                return

        # HANDSHAKE message (168 bytes)
        elif len(data) == 168:
            parsed = parse_handshake_message(data)
            if parsed:
                timestamp, ephemeral_pub, ephemeral_sig, pfs_hash = parsed
                self.state["handshake_received"] = True
                index = len(self.inbound_messages)
                msg = {
                    "type": "HANDSHAKE",
                    "raw": data,
                    "parsed": (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash),
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                # (Optional auto-responder for handshake could go here)
                return

        # Check if the message might be an encrypted message (e.g. header of 18 bytes at start)
        elif len(data) > 18:
            # Try to parse header from encryption module
            try:
                from encryption import MessageHeader
                header = MessageHeader.unpack(data[:18])
                # If header unpacking is successful and data length fits header.data_len + header size + tag size:
                expected_len = 18 + header.data_len + 16  # tag is 16 bytes
                if len(data) == expected_len:
                    index = len(self.inbound_messages)
                    msg = {
                        "type": "ENCRYPTED_MESSAGE",
                        "raw": data,
                        "parsed": header,  # we can store header for further processing
                        "connection": conn
                    }
                    self.inbound_messages.append(msg)
                    print(f"{YELLOW}[NOTICE]{RESET} Stored inbound ENCRYPTED_MESSAGE at index={index}.")
                    return
            except Exception:
                pass

        # Otherwise, unrecognized/malformed message.
        index = len(self.inbound_messages)
        msg = {
            "type": "UNKNOWN",
            "raw": data,
            "parsed": None,
            "connection": conn
        }
        self.inbound_messages.append(msg)
        print(f"{RED}[WARNING]{RESET} Unrecognized or malformed message stored at index={index}.")


        # -------------------------------------------------------------------------
        # HKDF Derivation
        # -------------------------------------------------------------------------

    def derive_hkdf(self):
        """
        Derives a 256-bit key using HKDF.
        Uses as input keying material (IKM) the shared secret from ECDH.
        The salt is computed as SHA256(session_nonce || pfs_param), where:
          - session_nonce is taken from self.session_nonce (32 bytes) or defaults to zeros.
          - pfs_param is taken from the first inbound HANDSHAKE's pfs_hash field (32 bytes) or zeros.
        """
        if not self.shared_secret:
            print(f"{RED}[ERROR]{RESET} No shared secret available; cannot derive HKDF key.")
            return

        # IKM: shared secret converted from hex to bytes.
        ikm = bytes.fromhex(self.shared_secret)
        # Use stored session_nonce if available; otherwise default to zeros.
        session_nonce = self.session_nonce if self.session_nonce is not None else (b"\x00" * 32)

        # Determine pfs_param from first HANDSHAKE message (if any)
        pfs_param = None
        for msg in self.inbound_messages:
            if msg["type"] == "HANDSHAKE":
                # Expect parsed handshake as tuple: (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash)
                try:
                    _, _, _, pfs_param = msg["parsed"]
                except Exception:
                    pfs_param = None
                break
        if pfs_param is None:
            print(f"{RED}[WARNING]{RESET} No HANDSHAKE found; using 32 zero bytes for pfs_param.")
            pfs_param = b"\x00" * 32  # 256-bit zeros

        # Ensure both are bytes
        if isinstance(session_nonce, str):
            session_nonce = session_nonce.encode()
        if isinstance(pfs_param, str):
            pfs_param = pfs_param.encode()

        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.kdf.hkdf import HKDF
        hasher = hashes.Hash(hashes.SHA256())
        hasher.update(session_nonce + pfs_param)
        salt_value = hasher.finalize()

        hkdf = HKDF(
            algorithm=hashes.SHA256(),
            length=32,  # 256 bits
            salt=salt_value,
            info=b"",
        )
        derived_key = hkdf.derive(ikm)
        self.hkdf_key = derived_key.hex()
        print(f"{GREEN}[HKDF]{RESET} Derived HKDF key: {self.hkdf_key}")

    # -------------------------------------------------------------------------
    # Auto-responder helpers
    # -------------------------------------------------------------------------

    def _auto_respond_ping(self, index: int):
        """
        Called by a Timer to respond automatically to a PING_REQUEST after 2s.
        """
        print(f"{BLUE}[AUTO]{RESET} Delayed responding to PING at index={index}")
        self.respond_to_ping(index, answer_code=0)
        self.show_state()

    def _auto_respond_handshake(self, index: int):
        """
        Called by a Timer to handle inbound HANDSHAKE automatically.
        1) Generate ephemeral keys if not already set
        2) Compute ECDH with the inbound ephemeral pub (generate_ecdhe)
        3) Send our handshake back
        4) Show state
        """
        print(f"{BLUE}[AUTO]{RESET} Delayed ECDH process for HANDSHAKE at index={index}")

        # 1) Generate ephemeral keys if we don't have them
        if not self.ephemeral_privkey or not self.ephemeral_pubkey:
            self.generate_ephemeral_keys()

        # 2) Compute ECDH from inbound ephemeral pub
        self.generate_ecdhe(index)

        # 3) Send our handshake to the peer
        self.send_handshake()

        # 4) Show final state
        self.show_state()

    # -------------------------------------------------------------------------
    # Public Methods
    # -------------------------------------------------------------------------

    def connect_to_peer(self, port: int):
        conn = transmission.connect_to_peer("127.0.0.1", port, self.on_data_received)
        self.connections.append(conn)
        print(f"{GREEN}[IcingProtocol]{RESET} Outgoing connection to port {port} established.")

    def set_peer_identity(self, peer_pubkey_hex: str):
        pubkey_bytes = bytes.fromhex(peer_pubkey_hex)
        self.peer_identity_pubkey_obj = load_peer_identity_key(pubkey_bytes)
        self.peer_identity_pubkey_bytes = pubkey_bytes
        print(f"{GREEN}[IcingProtocol]{RESET} Stored peer identity pubkey (hex={peer_pubkey_hex[:16]}...).")

    def generate_ephemeral_keys(self):
        self.ephemeral_privkey, self.ephemeral_pubkey = get_ephemeral_keypair()
        print(f"{GREEN}[IcingProtocol]{RESET} Generated ephemeral key pair: pubkey={self.ephemeral_pubkey.hex()[:16]}...")

    # Updated send_ping_request: generate a 17-byte nonce, extract top 129 bits, and store that (as bytes)
    def send_ping_request(self):
        if not self.connections:
            print(f"{RED}[ERROR]{RESET} No active connections.")
            return
        nonce_full = os.urandom(17)
        # Compute the 129-bit session nonce: take the top 129 bits.
        nonce_int = int.from_bytes(nonce_full, 'big') >> 7  # 129 bits
        session_nonce_bytes = nonce_int.to_bytes(17, 'big')  # still 17 bytes but only 129 bits are meaningful
        if self.session_nonce is None:
            self.session_nonce = session_nonce_bytes
            print(f"{YELLOW}[NOTICE]{RESET} Stored session nonce from sent PING.")
        pkt = build_ping_request(version=0, cipher=0, nonce_full=nonce_full)
        self._send_packet(self.connections[0], pkt, "PING_REQUEST")
        self.state["ping_sent"] = True

    def send_handshake(self):
        """
        Build and send handshake:
          - 32-bit timestamp
          - ephemeral_pubkey (64 bytes, raw x||y)
          - ephemeral_signature (64 bytes, raw r||s)
          - pfs_hash (32 bytes)
          - 32-bit CRC
        """
        if not self.connections:
            print(f"{RED}[ERROR]{RESET} No active connections.")
            return
        if not self.ephemeral_privkey or not self.ephemeral_pubkey:
            print(f"{RED}[ERROR]{RESET} Ephemeral keys not generated.")
            return
        if self.peer_identity_pubkey_bytes is None:
            print(f"{RED}[ERROR]{RESET} Peer identity not set; needed for PFS tracking.")
            return

        # 1) Sign ephemeral_pubkey as r||s
        #    Instead of DER, we do raw r||s each 32 bytes
        sig_der = sign_data(self.identity_privkey, self.ephemeral_pubkey)
        # Convert DER -> (r, s) -> raw 64 bytes
        # Quick approach to parse DER using cryptography, or do a custom parse
        from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature
        r_int, s_int = decode_dss_signature(sig_der)
        r_bytes = r_int.to_bytes(32, 'big')
        s_bytes = s_int.to_bytes(32, 'big')
        raw_signature = r_bytes + s_bytes  # 64 bytes

        # 2) PFS hash
        session_number, last_secret_hex = self.pfs_history.get(self.peer_identity_pubkey_bytes, (-1, ""))
        pfs = compute_pfs_hash(session_number, last_secret_hex)

        # 3) Build handshake
        timestamp_32 = int(time.time() * 1000) & 0xffffffff
        pkt = build_handshake_message(
            timestamp_32,
            self.ephemeral_pubkey,  # 64 bytes raw
            raw_signature,  # 64 bytes raw
            pfs  # 32 bytes
        )

        # 4) Send
        self._send_packet(self.connections[0], pkt, "HANDSHAKE")
        self.state["handshake_sent"] = True

    def enable_auto_responder(self, enable: bool):
        self.auto_responder = enable
        print(f"{BLUE}[AUTO]{RESET} Auto responder set to {enable}.")

    # -------------------------------------------------------------------------
    # Manual Responses
    # -------------------------------------------------------------------------

    def respond_to_ping(self, index: int, answer: int):
        if index < 0 or index >= len(self.inbound_messages):
            print(f"{RED}[ERROR]{RESET} Invalid index {index}.")
            return
        msg = self.inbound_messages[index]
        if msg["type"] != "PING_REQUEST":
            print(f"{RED}[ERROR]{RESET} inbound_messages[{index}] is not a PING_REQUEST.")
            return

        version = msg["parsed"]["version"]
        cipher = msg["parsed"]["cipher"]
        # Force cipher to 0 if it's not 0 (only AES-256-GCM is supported)
        if cipher != 0:
            print(
                f"{YELLOW}[NOTICE]{RESET} Received PING with unsupported cipher ({cipher}); forcing cipher to 0 in response.")
            cipher = 0

        conn = msg["connection"]
        resp = build_ping_response(version, cipher, answer)
        self._send_packet(conn, resp, "PING_RESPONSE")
        print(f"{BLUE}[MANUAL]{RESET} Responded to ping with answer={answer}.")

    def generate_ecdhe(self, index: int):
        """
        Formerly 'respond_to_handshake'. Verifies the inbound ephemeral signature
        and computes the ECDH shared secret, updating PFS history.
        """
        if index < 0 or index >= len(self.inbound_messages):
            print(f"{RED}[ERROR]{RESET} Invalid index {index}.")
            return
        msg = self.inbound_messages[index]
        if msg["type"] != "HANDSHAKE":
            print(f"{RED}[ERROR]{RESET} inbound_messages[{index}] is not a HANDSHAKE.")
            return

        # Unpack the tuple directly:
        timestamp, ephemeral_pub, ephemeral_sig, pfs_hash = msg["parsed"]

        # Use our raw_signature_to_der wrapper only if signature is 64 bytes.
        # Otherwise, assume the signature is already DER-encoded.
        from crypto_utils import raw_signature_to_der
        if len(ephemeral_sig) == 64:
            sig_der = raw_signature_to_der(ephemeral_sig)
        else:
            sig_der = ephemeral_sig

        ok = verify_signature(self.peer_identity_pubkey_obj, sig_der, ephemeral_pub)
        if not ok:
            print(f"{RED}[ERROR]{RESET} Ephemeral signature invalid.")
            return
        print(f"{GREEN}[OK]{RESET} Ephemeral signature verified.")

        if not self.ephemeral_privkey:
            print(f"{YELLOW}[WARN]{RESET} No ephemeral_privkey available, cannot compute shared secret.")
            return
        shared = compute_ecdh_shared_key(self.ephemeral_privkey, ephemeral_pub)
        self.shared_secret = shared.hex()
        print(f"{GREEN}[OK]{RESET} Computed ECDH shared key = {self.shared_secret}")

        old_session, _ = self.pfs_history.get(self.peer_identity_pubkey_bytes, (-1, ""))
        new_session = 1 if old_session < 0 else old_session + 1
        self.pfs_history[self.peer_identity_pubkey_bytes] = (new_session, self.shared_secret)

    # -------------------------------------------------------------------------
    # Utility
    # -------------------------------------------------------------------------

    def _send_packet(self, conn: transmission.PeerConnection, data: bytes, label: str):
        bits_count = len(data) * 8
        print(f"{BLUE}[SEND]{RESET} {label} -> {bits_count} bits: {data.hex()[:60]}{'...' if len(data.hex())>60 else ''}")
        conn.send(data)

    def show_state(self):
        print(f"\n{YELLOW}=== Global State ==={RESET}")
        print(f"Listening Port: {self.local_port}")
        print(f"Identity PubKey: 512 bits => {self.identity_pubkey.hex()[:16]}...")

        if self.peer_identity_pubkey_bytes:
            print(f"Peer Identity PubKey: 512 bits => {self.peer_identity_pubkey_bytes.hex()[:16]}...")
        else:
            print("Peer Identity PubKey: [None]")

        print("\nEphemeral Keys:")
        if self.ephemeral_pubkey:
            print(f"  ephemeral_pubkey: 512 bits => {self.ephemeral_pubkey.hex()[:16]}...")
        else:
            print("  ephemeral_pubkey: [None]")

        print(f"\nShared Secret: {self.shared_secret if self.shared_secret else '[None]'}")

        if self.hkdf_key:
            print(f"HKDF Derived Key: {self.hkdf_key} (size: {len(self.hkdf_key)*8} bits)")
        else:
            print("HKDF Derived Key: [None]")
        if self.session_nonce:
            # session_nonce now contains 17 bytes, but only the top 129 bits are used.
            print(f"Session Nonce: {self.session_nonce.hex()} (129 bits)")
        else:
            print("Session Nonce: [None]")
        print("\nProtocol Flags:")
        for k, v in self.state.items():
            print(f"  {k}: {v}")

        print("\nAuto Responder:", self.auto_responder)

        print("\nActive Connections:")
        for i, c in enumerate(self.connections):
            print(f"  [{i}] Alive={c.alive}")

        print("\nInbound Message Queue:")
        for i, m in enumerate(self.inbound_messages):
            print(f"  [{i}] type={m['type']} len={len(m['raw'])} bytes => {len(m['raw']) * 8} bits")
        print()

    def stop(self):
        self.server_listener.stop()
        for c in self.connections:
            c.close()
        self.connections.clear()
        self.inbound_messages.clear()
        print(f"{RED}[STOP]{RESET} Protocol stopped.")

    # New method: Send an encrypted message over the first active connection.
    def send_encrypted_message(self, plaintext: str):
        """
        Encrypts the provided plaintext (a UTF-8 string) using the derived HKDF key (AES-256),
        and sends the encrypted message over the first active connection.
        The message format is: header (18 bytes) || ciphertext || tag (16 bytes).
        """
        if not self.connections:
            print(f"{RED}[ERROR]{RESET} No active connections.")
            return
        if not self.hkdf_key:
            print(f"{RED}[ERROR]{RESET} No HKDF key derived. Cannot encrypt message.")
            return
        key = bytes.fromhex(self.hkdf_key)
        plaintext_bytes = plaintext.encode('utf-8')
        encrypted = encrypt_message(plaintext_bytes, key)
        # Send the encrypted message over the first connection.
        self._send_packet(self.connections[0], encrypted, "ENCRYPTED_MESSAGE")
        print(f"{GREEN}[SEND_ENCRYPTED]{RESET} Encrypted message sent.")

    # New method: Decrypt an encrypted message provided as a hex string.
# New command: decrypt a received encrypted message from the inbound queue.
    def decrypt_received_message(self, index: int):
        if index < 0 or index >= len(self.inbound_messages):
            print(f"{RED}[ERROR]{RESET} Invalid message index.")
            return
        msg = self.inbound_messages[index]
        # Expect the message to be an encrypted transmission (we assume it is in the proper format).
        encrypted = msg["raw"]
        if not self.hkdf_key:
            print(f"{RED}[ERROR]{RESET} No HKDF key derived. Cannot decrypt message.")
            return
        key = bytes.fromhex(self.hkdf_key)
        try:
            plaintext_bytes = decrypt_message(encrypted, key)
            plaintext = plaintext_bytes.decode('utf-8')
            print(f"{GREEN}[DECRYPTED]{RESET} Decrypted message: {plaintext}")
            return plaintext
        except Exception as e:
            print(f"{RED}[ERROR]{RESET} Decryption failed: {e}")
